#!/usr/bin/python

#####################################
# Gyroscope Controller
#####################################
# Author: Aron Schneider and Sophia Donderer

import smbus
import math
import time
import random
import log

class GyroMaster():

    def __init__(self, krull):

        # ******* SETTINGS ********** #
        self.minForce = 1000 # careful handling
        self.midForce = 10000 # common handling
        self.maxForce = 25000 # throwing / agitating
        self.notMovedTrigger = 5 # trigger if not moved for x seconds
        self.turnDelay = 3 # if turning is solved, wait x seconds till next check

        # ******* SYSTEM ********** #
        self.sideLeft = 6
        self.sideRight = 1
        self.sideBack = 4
        self.sideBelly = 2
        self.sideHead = 5

        self.logtype = "interaction"
        self.logcat = "gyro"

        self.krull = krull
        self.power_mgmt_1 = 0x6b
        self.power_mgmt_2 = 0x6c
        self.bus = smbus.SMBus(1) #revision board!
        self.address = 0x68
        self.oldX = 0
        self.oldY = 0
        self.oldZ = 0
        self.readInterval = 0.1
        self.readTimer = 0
        self.bus.write_byte_data(self.address, self.power_mgmt_1,0)

        self.lastMoved = 0
        self.heldStill = False
        self.currentSide = 0
        self.turnedTimer = 0

        self.looping = True

    def getRandomSide(self):
        return random.choice([self.sideLeft, self.sideRight, self.sideBack, self.sideHead])

    def read_byte(self, adr):
        return self.bus.read_byte_data(self.address, adr)

    def read_word(self, adr):
        high = self.bus.read_byte_data(self.address, adr)
        low = self.bus.read_byte_data(self.address, adr+1)
        val = (high << 8) + low
        return val

    def read_word_2c(self, adr):
        val = self.read_word(adr)
        if(val >= 0x8000):
            return -((65535-val)+1)
        else:
            return val

    def dist(self, a,b):
        return math.sqrt((a*a)+(b*b))

    def get_y_rotation(self, x,y,z):
        radians = math.atan2(x, self.dist(y,z))
        return -math.degrees(radians)

    def get_x_rotation(self, x,y,z):
        radians= math.atan2(y,self.dist(x,z))
        return math.degrees(radians)

    def get_z_rotation(self, x,y,z):
        radians= math.atan2(z,self.dist(x,y))
        return math.degrees(radians)

    def loop(self):
        log.debug("gyro loop started")

        while self.looping:

            accel_xout = self.read_word_2c(0x3b)
            accel_yout = self.read_word_2c(0x3d)
            accel_zout = self.read_word_2c(0x3f)

            self.detectMovement(accel_xout, accel_yout, accel_zout)

            accel_xout_scaled = accel_xout/16384.0
            accel_yout_scaled = accel_yout/16384.0
            accel_zout_scaled = accel_zout/16384.0

            x_rot = self.get_x_rotation(accel_xout_scaled, accel_yout_scaled, accel_zout_scaled)
            y_rot = self.get_y_rotation(accel_xout_scaled, accel_yout_scaled, accel_zout_scaled)
            z_rot = self.get_z_rotation(accel_xout_scaled, accel_yout_scaled, accel_zout_scaled)

            self.detectTurn(x_rot, y_rot, z_rot)

            time.sleep(0.01)


    def detectMovement(self, newX, newY, newZ):

        # if no movement set
        if self.oldX == 0: # start
            self.oldX = newX
            self.oldY = newY
            self.oldZ = newZ
            self.readTimer = time.time()
        else:
            t = time.time() - self.readTimer
            # calculate only every x seconds
            if t >= self.readInterval:
                force = math.sqrt((newX-self.oldX)*(newX-self.oldX) + (newY-self.oldY)*(newY-self.oldY) + (newZ-self.oldZ )*(newZ-self.oldZ ))
                # Krull is moved
                if force >= self.midForce:
                    self.lastMoved = time.time()
                    self.heldStill = False
                # if not moved for x seconds:
                elif time.time() - self.lastMoved >= self.notMovedTrigger:
                    self.heldStill = True
                if force >= self.maxForce:
                    log.log(self.logtype, self.logcat, "throw")
                    self.krull.thrown()
                    pass
                self.oldX = newX
                self.oldY = newY
                self.oldZ = newZ

    def detectTurn(self, x_rot, y_rot, z_rot):

        t = time.time() - self.turnedTimer
        if t >= self.turnDelay:
            side = 0

            if x_rot < 20 and x_rot > -20 and y_rot > 70 and z_rot < 20 and z_rot > -20:
                side = 1
            if x_rot < 20 and x_rot > -20 and y_rot < 20 and y_rot > -20 and z_rot > 70:
                side = 2
            if x_rot > 70 and y_rot < 20 and y_rot > -20 and z_rot <20 and z_rot >-20:
                side = 3
            if x_rot < 20 and x_rot > -20 and y_rot < 20 and y_rot > -20 and z_rot < -70:
                side = 4
            if x_rot < -70 and y_rot < 20 and y_rot > -20 and z_rot < 20 and z_rot > -20:
                side = 5
            if x_rot < 20 and x_rot > -20 and y_rot < -70 and z_rot <20 and z_rot > -20:
                side = 6

            if side != self.currentSide and side != 0:
                log.log(self.logtype, self.logcat, "turned on side "+str(side))
                self.krull.turned(side)
                self.currentSide = side

				
