#!/usr/bin/python

#####################################
# Karsten Krull MainController
# Author: Aron Schneider
#####################################

import time
from threading import *
import random
import RPi.GPIO as GPIO
import multiprocessing

#Krull classes
import log
import state
import speaker
import player
import mic
import light
import matrix
import teeth
import rfid
import gyro

class Krull():
    def __init__(self):

        # *********** SETTINGS ************** #
        self.generalDelay = 1
        self.registerDelay = 5 # dealy betweeen registerRequest
        self.eatDelay = 3 # delay before register next food
        self.drinkDelay = 3 # delay before register next food
        self.throwDelay = 3 # delay before register next throwing
        self.turnDelay = 3 # delay before register next throwing
        self.shoutDelay = 5 # delay before register next shout
        self.singDelay = 3 # delay before register next sing

        self.updateInterval = 1 # Interval between condition updates
        self.randomInterval = 10 # Triggerinterval for random condition

        self.mood = 5  # Mainstate of Krull. If (mood == 0): game lost
        self.moodDropChance = 25 # 30% chance to lower mood
        self.moodDropActivateLevel = 3 # amount of Conditions in emergency level to activate mood drop

        self.shoutIntervalMin = 5 # minimal start interval betweeen shouting/whining/requests
        self.shoutIntervalMax = 8 # maximal start interval betweeen shouting/whining/requests
        self.shoutIntervalSub = dict(level0 = 0, level1 = 1, level2 = 2, level3 = 4) # subtract from shoutInterval if on emergency Levels
        self.shoutIntervalHeadache = 12

        self.hickupInterval = 5 # if hickup interval between hickupshouts

        self.eyeFeedbackBlinks = 5 # amount of on/offs
        self.eyeFeedbackDuration = 0.3 # light on
        self.eyeFeedbackInterval = 0.4 # light off


        # Deactivate Inputs
        self.deactivateMic = False
        self.deactivateLight = False
        self.deactivateScale = False
        self.deactivateTeeth = False

        # ******* SYSTEM ********** #

        self.threadsRunning = False
        self.led = "none"

        self.setupLock = True
        self.systemSetup()

        self.gamestate = "start" # start brain end restart idle



    # ******* SYSTEM ********** #
    def systemSetup(self):

        self.setupLock = True

        log.start()
        log.log("game", "state", "system setup")

        # init mainvalues
        self.shoutInterval = 0
        self.brainLoop = False
        self.idling = False

        self.conditions = dict(
            normal = state.State(self, "normal"),
            hungry = state.StateActive(self, "hungry"),
            thirsty = state.StateActive(self, "thirsty"),
            bored = state.StateBored(self, "bored"),
            headache = state.StateHeadache(self, "headache"),
            sick = state.StateSick(self, "sick"),
            toothache = state.StateToothache(self, "toothache"),
            scale = state.StateScale(self, "scale"),
            pain = state.StatePain(self, "pain"),
            hickup = state.StateHickup(self, "hickup"),
        )
        self.mainCondition = "normal"

        self.newPlayerRequest = False
        self.complaining = False
        self.setShoutInvterval()
        self.emergencyLevel = 0

        self.turnTimer = 0
        self.updateTimer = 0
        self.complainTimer = 0
        self.eatTimer = 0
        self.drinkTimer = 0
        self.throwTimer = 0
        self.randomTimer = 0
        self.hickupTimer = 0
        self.shoutTimer = 0
        self.singTimer = 0

        self.complainLock = False
        self.shoutLock = False
        self.playerRequestLevel = 1
        self.currentComplaint = self.conditions["normal"]

         # init objects
        self.mouth = speaker.Speaker()
        self.player = player.PlayerController(self)

        # start input controll threads
        if self.led == "none":
            q = multiprocessing.Queue()
            self.led = matrix.MatrixController(q, args=("ledmatrix",))
            self.led.start()

        self.rfid = rfid.RFID(self)
        self.rfid = rfid.RFID(self)
        self.threadRfid = Thread(target=self.rfid.loop)
        self.threadRfid.setDaemon(True)
        self.threadRfid.start()

        self.teeth = teeth.Teeth(self)
        self.threadTeeth = Thread(target=self.teeth.loop)
        self.threadTeeth.setDaemon(True)
        self.threadTeeth.start()


        self.setupLock = False



    # Gameloop
    def update(self):
        try:
            while True:
                if self.gamestate == "start":
                    self.start()
                elif self.gamestate == "register":
                    self.register()
                elif self.gamestate == "initBrain":
                    self.initBrain()
                elif self.gamestate == "brain":
                    self.brain()
                elif self.gamestate == "restart":
                    self.restart()
                elif self.gamestate == "end":
                    self.end()
                else:
                    self.idle()
                time.sleep(0.01)
        except Exception as e:
            log.exception()
        except KeyboardInterrupt:
            pass
        finally:
            GPIO.cleanup()

    # Start
    def start(self):

        log.debug("start loop: systemlock = "+str(self.setupLock))

        if self.setupLock == False:

            log.log("game", "state", "start")

            self.eyeFeedback()

            # play random wakeup line
            self.mouth.speak("game_wakeup", True)
            self.delay()

            # play random reaction to players line
            self.mouth.speak("game_firstreaction")
            self.delay()

            # play random explanation line
            self.mouth.speak("game_explanation")
            self.delay()

            # play random register line
            self.mouth.speak("login_request")
            self.delay()

            self.newPlayerRequest = True

            self.startTimer = time.time()
            log.log("game", "state", "register")
            self.setGamestate("register")

    def register(self):
        # waite till first player logged in
        if not self.player.loginLock:
            # every x-y seconds play random login line
            t = time.time() - self.startTimer
            if t >= self.registerDelay :
                self.mouth.speak("login_request")
                self.delay()
                self.startTimer = time.time()


    def initBrain(self):
        log.log("game", "state", "init brain")
        self.startThreads()
        self.updateTimer = time.time()
        self.complainTimer = time.time()
        self.shoutInterval = random.randrange(self.shoutIntervalMin, self.shoutIntervalMax)

        # Start Brain
        self.setGamestate("brain")
        log.log("game", "state", "start brain")

    def startThreads(self):
        log.log("game", "state", "start threads")
        self.threadsRunning = True

        self.gyro = gyro.GyroMaster(self)
        self.threadGyro = Thread(target=self.gyro.loop)
        self.threadGyro.setDaemon(True)
        self.threadGyro.start()

        self.mic = mic.Microphone(self)

        self.light = light.Light(self)
        self.threadLight = Thread(target=self.light.loop)
        self.threadLight.setDaemon(True)
        self.threadLight.start()

        if not self.deactivateScale:
            import scales
            self.scale = scales.Scales(self)
        bored = self.conditions.get("bored")
        self.boredChecked= Thread(target=bored.loop)
        self.boredChecked.setDaemon(True)
        self.boredChecked.start()

        self.teeth.puzzle = True

    # Brain loop, controlling Krulls mood, condition and reactions (audioOutput)
    def brain(self):
        # Update each Condition every x tick
        t = time.time() - self.updateTimer
        if t >= self.updateInterval :
            self.emergencyLevel = 0
            for key, value in self.conditions.iteritems():
                value.update()
                if value.getEmergency():
                    self.emergencyLevel += 1
            # check if more than 3 conditions are in emergency state
            if self.emergencyLevel >= self.moodDropActivateLevel:
                self.emergencyLevel = self.moodDropActivateLevel
                self.emergency()
            self.updateTimer = time.time()

        if self.gamestate == "brain":
            # trigger random states
            self.randomTrigger()

            # hickup
            if self.conditions.get("hickup").level > 0:
                self.hickup()
                time.sleep(3)

            #complain
            if not self.complainLock:
                self.complain()


    def randomTrigger(self):
        # Trigger random-conditions
        t = time.time() - self.randomTimer
        if t >= self.randomInterval:
            #log.log("condition:", "trigger random")
            if not self.conditions.get("pain").active:
                self.conditions.get("pain").trigger()
            if not self.conditions.get("scale").active and not self.deactivateScale:
                self.conditions.get("scale").trigger()
            self.randomTimer = time.time()

    def hickup(self):
        if self.hickupTimer == 0:
            self.hickupTimer = time.time()
        else:
            t = time.time() - self.hickupTimer
            if t >= self.hickupInterval:
                self.mouth.speak("complaining_hickup")

    def complain(self):
        # Krull Complaining
        if self.complaining:
            if not self.mouth.speaking():
                self.complaining = False
                self.complainTimer = time.time()
                self.shoutInterval = random.randrange(self.shoutIntervalMin, self.shoutIntervalMax)
        else:
            if self.currentComplaint.name == "headache":
                interval = self.shoutIntervalHeadache
            else:
                interval = self.shoutInterval - self.shoutIntervalSub["level"+str(self.emergencyLevel)]

            t = time.time() - self.complainTimer
            if t >= interval:

                # if Krull wants to go to another player
                if self.newPlayerRequest:
                    n = self.currentComplaint.name
                    if n == "bored":
                        line = "handover_bored"
                    elif n == "headache":
                        line = "handover_headache"
                    elif n == "sick":
                        line = "handover_sick"
                    elif n == "hickup":
                        line = "handover_hickup"
                    else:
                        if self.playerRequestLevel == 1:
                            line = "handover_normal"
                        else:
                            line = "handover_reminder"
                    self.mouth.speak(line)
                    self.complaining = True
                    if self.playerRequestLevel == 1:
                        self.playerRequestLevel = 2

                # if problem not solved
                else:
                    priorityComplaint = self.conditions["normal"]
                    highestLevel = 1

                    for key, value in self.conditions.iteritems():
                        if value.active == True:
                            log.debug("active condition: "+value.name)
                            level = value.level
                            if level > highestLevel:
                                #log.log("debug", "new highest condition: "+value.name)
                                priorityComplaint = value
                                highestLevel = level
                            elif level == highestLevel:
                                #log.log("debug", "equal condition level: "+value.name)
                                # if there is no  priority complaint assigned, assign current
                                if priorityComplaint.name == "normal":
                                    priorityComplaint = value
                                    #log.log("debug", "old priority = normal "+value.name+"  assigned as new priority condition. ")
                                # assign new complaint if its level timer is higher than leveltimer of current  priority complaint
                                elif value.levelTimer > priorityComplaint.levelTimer:
                                    priorityComplaint = value
                                    #log.log("debug", value.name+" has higher levelTimer. -> Assigned as new priority condition ")
                                highestLevel = level
                            else:
                                #log.log("debug", priorityComplaint.name+" remains priority")
                                pass
                    self.currentComplaint = priorityComplaint
                    log.log("game", "state", "current complaint: "+self.currentComplaint.name)

                    # if pain check side
                    if self.currentComplaint.name == "pain":
                        side = self.conditions.get("pain").side
                        #log.log("complaining", "Side of Pain: "+str(side))
                        line = "complaining_"+self.currentComplaint.name
                        if side == self.gyro.sideLeft:
                            line += "Left"
                        elif side == self.gyro.sideRight:
                            line +=  "Right"
                        elif side == self.gyro.sideBack:
                            line += "Back"
                        elif side == self.gyro.sideHead:
                            line += "Head"
                        else:
                            log.warning("No side of pain found. Side: "+str(side))
                    # if headache check for light or noise
                    elif self.currentComplaint.name == "headache":
                        line = "complaining_"+self.currentComplaint.name
                         # light environment
                        if not self.deactivateLight and self.light.current == self.light.light:
                            line += "Light"
                        # noise environment
                        elif self.mic.isNoisy():
                            line += "Noisy"
                        line += str(highestLevel)
                    # if bored check for bored task
                    elif self.currentComplaint.name == "bored":
                        line = "complaining_"+self.currentComplaint.name
                        task = self.conditions.get("bored").task
                        if task == "sing":
                            line += "Sing"
                        elif task == "throw":
                            line += "Throw"
                        elif not self.deactivateLight and task == "light":
                            line += "Light"
                        else:
                            line += "Any"
                    elif self.currentComplaint.name == "normal":
                        line = "complaining_"+self.currentComplaint.name
                    else:
                        line = "complaining_"+self.currentComplaint.name+str(highestLevel)
                    self.mouth.speak(line)
                    self.delay()
                    self.complaining = True

    def setShoutInvterval(self):
        self.shoutInterval = random.randrange(self.shoutIntervalMin, self.shoutIntervalMax)

    def playerLogin(self, uid):
        self.complainLock = True
        self.player.playerLogin(uid)
        self.complainLock = False

    def eat(self, uid):
        #log.log("con", "Eat")
        if self.gamestate == "brain":
            t = time.time() - self.eatTimer
            if t >= self.eatDelay:
                self.complainLock = True
                # Reaction to food if hunger is current highest condition
                if self.currentComplaint.name == "hungry":
                    self.eatSound("reaction_munching_hungry", "reaction_food_hungry")
                # Reactions if he is not hungry:
                elif self.currentComplaint.name == "normal":
                    self.eatSound("reaction_munching_normal", "reaction_food_normal")
                elif self.currentComplaint.name == "bored":
                    self.eatSound("reaction_munching_normal", "reaction_food_bored")
                elif self.currentComplaint.name == "sick":
                    self.eatSound("reaction_munching_normal", "reaction_food_sick")
                elif self.currentComplaint.name == "toothache":
                    self.eatSound("reaction_munching_normal", "reaction_food_toothache")
                elif self.currentComplaint.name == "thirsty":
                    self.eatSound("reaction_munching_hungry", "reaction_food_thirsty")
                elif self.currentComplaint.name == "hickup":
                    self.eatSound("reaction_munching_hickup", "reaction_food_hickup")
                # if there is no special reaction to food react according to hunger-level
                else:
                    if self.conditions.get("hungry").level > 0:
                        self.eatSound("reaction_munching_hungry", "reaction_food_hungry")
                    else:
                        self.eatSound("reaction_munching_normal", "reaction_food_normal")

                # if hungry eating solves hungry-condition, even if other conditions have a higher priority
                if self.conditions.get("hungry").level > 0:
                    #log.log("Condition", "solved hungry")
                    self.conditions.get("hungry").solve()
                    self.solve()

                # giving food can always trigger negative effects
                self.conditions.get("hickup").trigger()
                self.conditions.get("sick").trigger()
                if not self.deactivateTeeth:
                    self.conditions.get("toothache").trigger()

                self.eatTimer = time.time()
                self.complainLock = False

    def solve(self):
        self.complainLock = True
        log.nextTurn()
        self.playerRequestLevel = 1
        self.newPlayerRequest = True
        self.currentComplaint = self.conditions["normal"]
        self.eyeFeedback()
        self.complainLock = False

    def eatSound(self, nomnom, reaction):
        self.mouth.speak(nomnom, True)
        while self.mouth.speaking():
            time.sleep(0.1)
        self.mouth.speak(reaction)

    def drink(self, uid):
        #log.log("Input", "Drink")
        if self.gamestate == "brain":
            t = time.time() - self.drinkTimer
            if t >= self.drinkDelay:
                self.complainLock = True
                # Current highest condition
                if self.currentComplaint.name == "thirsty":
                    self.eatSound("reaction_gulp_normal", "reaction_drink_thirsty")
                # Special reactions:
                elif self.currentComplaint.name == "normal":
                    self.eatSound("reaction_gulp_normal", "reaction_drink_normal")
                elif self.currentComplaint.name == "bored":
                    self.eatSound("reaction_gulp_normal", "reaction_drink_bored")
                elif self.currentComplaint.name == "sick":
                    self.eatSound("reaction_gulp_normal", "reaction_drink_sick")
                elif self.currentComplaint.name == "toothache":
                    self.eatSound("reaction_gulp_normal", "reaction_drink_toothache")
                elif self.currentComplaint.name == "hungry":
                    self.eatSound("reaction_gulp_normal", "reaction_drink_hungry")
                elif self.currentComplaint.name == "hickup":
                    self.eatSound("reaction_gulp_hickup", "reaction_drink_hickup")
                # no special reaction
                else:
                    if self.conditions.get("thirsty").level > 0:
                        self.eatSound("reaction_gulp_normal", "reaction_drink_thirsty")
                    else:
                        self.eatSound("reaction_gulp_normal", "reaction_drink_normal")

                # positive effect
                if self.conditions.get("thirsty").level > 0:
                    #log.log("Condition", "solved thirsty")
                    self.conditions.get("thirsty").solve()
                    self.solve()

                # negative effect
                self.conditions.get("hickup").trigger()
                self.conditions.get("sick").trigger()
                if not self.deactivateTeeth:
                    self.conditions.get("toothache").trigger()

                self.drinkTimer = time.time()
                self.complainLock = False

    # action if Krull is moved hard or thrown
    def thrown(self):
        #log.log("Interaction", "Thrown")
        t = time.time() - self.throwTimer
        if t >= self.throwDelay:
            self.complainLock = True
            # Special reactions:
            if self.currentComplaint.name == "bored":
                self.mouth.speak("reaction_throw_bored", True)
            elif self.currentComplaint.name == "headache":
                self.mouth.speak("reaction_throw_headache", True)
            elif self.currentComplaint.name == "hickup":
                self.mouth.speak("reaction_throw_hickup", True)
            elif self.currentComplaint.name == "sick":
                self.mouth.speak("reaction_throw_sick", True)
            # no special reaction
            else:
                self.mouth.speak("reaction_throw_normal", True)

            # positive effect
            if self.conditions.get("bored").level > 0:
                #log.log("Condition", "solved bored (throw)")
                self.conditions.get("bored").solve("throw")
                self.mouth.speak("reaction_throw_bored", True)

            # negative effect
            #self.conditions.get("headache").trigger()
            self.conditions.get("sick").trigger()

            self.throwTimer = time.time()
            self.complainLock = False

    # action if Krull is turned on a new side
    def turned(self, side):
        #log.log("Input", "Turned")
        t = time.time() - self.turnTimer
        if t >= self.turnDelay:
            self.complainLock = True
            pain = self.conditions.get("pain")
            if pain.level >=1:
                if pain.side == side:
                    #log.log("Condition", "solved pain")
                    self.gyro.turnedTimer = time.time()
                    pain.solve()
                    self.solve()
                    self.mouth.speak("reaction_turned_right")
                    self.turnTimer = time.time()
                elif pain.side != self.gyro.sideBack:
                    log.log("Interaction", "gyro", "Turned on wrong side")
                    self.mouth.speak("reaction_turned_wrong")
            self.complainLock = False

    def sing(self):
        t = time.time() - self.singTimer
        if t >= self.singDelay:
            #log.log("Input", "Singing")
            self.complainLock = True
            # positive effect
            if self.conditions.get("bored").level > 0:
                #log.log("Condition", "solved bored (sing)")
                self.conditions.get("bored").solve("sing")
                self.mouth.speak("reaction_sing")
            # negative effect
            self.conditions.get("headache").trigger()
            self.singTimer = time.time()
            self.complainLock = False

    def shout(self):
        #log.log("Input", "Shout")
        t = time.time() - self.shoutTimer
        if t >= self.shoutDelay:
            self.complainLock = True

            # positive effect
            if self.conditions.get("hickup").level > 0:
                self.solve()
                #log.log("Condition", "solved shout")
                self.conditions.get("hickup").solve()
                self.stopMic()
                self.mouth.speak("reaction_hickup_solved")
                #self.delay()
            elif self.conditions.get("bored").level > 0:
                self.conditions.get("bored").solve("sing")
                self.mouth.speak("reaction_sing")
            else:
                self.mouth.speak("reaction_shout")
                #self.delay()
                pass

            # negative effect
            self.conditions.get("headache").trigger()

            self.shoutTimer = time.time()
            self.complainLock = False

    def teethTouch(self):
        self.complainLock = True
        #log.log("input", "teeth touched")
        self.mouth.speak("reaction_teeth", True)
        self.complainLock = False

    def teethSolved(self):
        self.complainLock = True
        #log.log("condition", "solved toothache")
        self.mouth.speak("reaction_teeth_solved", True)
        self.conditions.get("toothache").solve()
        self.solve()
        self.complainLock = False

    def scaleTouch(self):
        self.complainLock = True
        #log.log("input", "scales touched")
        self.mouth.speak("reaction_scale", True)
        self.complainLock = False

    def scaleSolved(self):
        self.complainLock = True
        #log.log("condition", "solved scales")
        self.mouth.speak("reaction_scale_solved", True)
        self.conditions.get("scale").solve()
        self.solve()
        self.complainLock = False

    def headacheSolved(self):
        self.complainLock = True
        #log.log("condition", "solved headache")
        self.mouth.speak("reaction_headache_solved", True)
        self.solve()
        self.stopMic()
        self.complainLock = False

    def sickSolved(self):
        self.complainLock = True
        #log.log("condition", "solved sick")
        self.mouth.speak("reaction_sick_solved", True)
        self.solve()
        self.complainLock = False

    def boredLightSolved(self):
        self.complainLock = True
        #print "bored task: solved"
        self.mouth.speak("reaction_light", True)
        self.boredSolved()
        self.complainLock = False

    def boredSolved(self):
        self.complainLock = True
        self.solve()
        self.mouth.speak("reaction_bored_solved", True)
        self.complainLock = False
        self.stopMic()

    def startMic(self):
        if not self.mic.looping:
            log.debug("Start Mic")
            self.mic.looping = True
            self.threadMic = Thread(target=self.mic.loop)
            self.threadMic.setDaemon(True)
            self.threadMic.start()

    def stopMic(self):
        if not self.conditions.get("hickup").active and not self.conditions.get("headache").active and (not self.conditions.get("bored").active and self.conditions.get("bored").task == "sing") :
            if not self.mic.looping:
                log.debug("Stop Mic")
                self.mic.looping = False
                self.mic.stop()

    def eyeFeedback(self):
        self.led.queue.put(["eyeFeedback"])
        '''
        i = 1
        while i <= self.eyeFeedbackBlinks:
            self.led.eyes(self.eyeFeedbackDuration)
            time.sleep(self.eyeFeedbackInterval)
            i += 1
        '''

    def delay(self):
        while self.mouth.speaking():
            time.sleep(0.1)
        time.sleep(self.generalDelay)
        pass

    def emergency(self):
        log.log("game", "mood", "emergency mode")
        if random.randrange(0, 100) < self.moodDropChance:
            self.mood -= 1
            log.log("game", "mood", "mood decreased: "+str(self.mood))
            if self.mood <= 0:
                self.complainLock = True
                log.log("game", "end", "GAME LOST")
                #self.end()
                self.setGamestate("end")


    # Finish Loop
    def end(self):
        self.led.queue.put(["party"])
        #self.brainLoop = False
        self.exitThreads()
        self.mouth.speak("game_end1")
        self.delay()
        self.mouth.speak("game_end2")
        self.delay()
        self.mouth.speak("game_end3")
        self.delay()
        self.mouth.speak("game_end4")
        self.delay()
        self.setGamestate("idle")


    def exitThreads(self):
        self.conditions.get("bored").looping = False
        self.conditions.get("headache").solveloop = False
        self.conditions.get("sick").solveloop = False
        self.gyro.looping = False
        self.rfid.looping = False
        self.mic.looping = False
        if not self.deactivateLight:
            self.light.looping = False
        if not self.deactivateScale:
            self.scale.playing = False
        if not self.deactivateTeeth:
            self.teeth.puzzle = False
        self.threadsRunning = False

    def idle(self):
        #log.log("game", "state",  "idle")
        pass

    def restart(self):
        log.log("game", "state", "restart")
        self.led.queue.put(["reset"])
        time.sleep(2)
        if self.threadsRunning:
            log.debug("exit threads")
            self.exitThreads()
        self.systemSetup()
        self.setGamestate("start")

    def setGamestate(self, s):
        self.gamestate = s

o = Krull()
o.update()



