# !/usr/bin/python

#####################################
# LED Matrix
# Author: Aron Schneider
#####################################

import RPi.GPIO as GPIO
import time
import log
from threading import *
import multiprocessing


class MatrixController(multiprocessing.Process):

	def __init__(self, queue, args=(), kwargs=None):

		log.debug("LED matrix init")

		# *********** SETTINGS ************** #
		self.eyeFeedbackBlinks = 5 # amount of on/offs
		self.eyeFeedbackDuration = 0.3 # light on
		self.eyeFeedbackInterval = 0.4 # light off

		self.partyDuration = 0.1
		self.partyInterval = 0.1
		self.partyTime = 40

		# ******* SYSTEM ********** #
		multiprocessing.Process.__init__(self)

		self.queue = queue
		self.daemon = True
		self.receive_messages = args[0]

		self.tooth = 0;
		self.scales = False;

		GPIO.setmode(GPIO.BCM)

		#GPIO Pins
		self.cathodes = [5, 6, 13, 19]
		self.anodes = [12, 16, 20, 21]

		#LEDs
		self.leds = {
			"tooth1" : LED(0,0),
			"tooth2" : LED(0,1),
			"tooth3" : LED(0,2),
			"tooth4" : LED(0,3),

			"scale1" : LED(1,0),
			"scale2" : LED(1,1),
			"scale3" : LED(1,2),
			"scale4" : LED(1,3),

			"scale5" : LED(2,0),
			"scale6" : LED(2,1),
			"scale7" : LED(2,2),
			"scale8" : LED(2,3),

			"ledX" : LED(3,0),
			"ledY" : LED(3,1),
			"eyeLeft" : LED(3,2),
			"eyeRight" : LED(3,3)
		}

		for cathode in self.cathodes:
			GPIO.setup(cathode, GPIO.OUT)
			GPIO.output(cathode, 0)

		for anode in self.anodes:
			GPIO.setup(anode, GPIO.OUT)
			GPIO.output(anode, 0)

		self.looping = True
		self.partyTimer = 0

		self.oldLED = self.leds["ledX"]

		log.debug("LED matrix init end")


	def eyes(self, duration):
		self.setLED(self.oldLED, 0)
		self.setLED(self.leds["eyeRight"], 1)
		self.setLED(self.leds["eyeLeft"], 1)
		timerEye = Timer(duration, self.eyesOff)
		timerEye.start()

	def on(self, led, duration):
		self.setLED(self.oldLED, 0)
		self.eyesOff()
		self.setLED(self.leds[led], 1)
		self.oldLED = self.leds[led]
		timer = Timer(duration, self.off, [led])
		timer.start()

	def onInfinite(self, led):
		self.setLED(self.oldLED, 0)
		self.eyesOff()
		self.setLED(self.leds[led], 1)
		self.oldLED = self.leds[led]

	def off(self, led):
		self.setLED(self.leds[led], 0)

	def eyesOff(self):
		self.setLED(self.leds["eyeRight"], 0)
		self.setLED(self.leds["eyeLeft"], 0)

		if self.tooth != 0:
			self.setLED(self.leds[self.tooth] , 1)

	def run(self):
		log.debug("run")
		self.loop()
		log.debug("run ende")

	def do_thing_with_message(self, message):
		if self.receive_messages:
			print multiprocessing.current_process().name, "Received {}".format(message)

	def loop(self):
		#connection
		log.debug("looping")
		while self.looping:
			self.tasks()
			for led in self.leds.itervalues():
				if led.state == 1:
					self.blink(led)
			time.sleep(0.001)

	def tasks(self):
		if not self.queue.empty():
			task = self.queue.get()
			if task[0] == "on":
				self.onOld(task[1])
			elif task[0] == "off":
				self.offOld(task[1])
			elif task[0] == "blink":
				log.debug("led blink: "+task[1])
				self.onOff(task[1], task[2])
			elif task[0] == "eyeFeedback":
				log.debug("led eyefeedback")
				self.eyeFeedback()
			elif task[0] == "reset":
				log.debug("led reset")
				self.reset()
			elif task[0] == "party":
				self.party()
			elif task[0] == "check":
				self.check()

	def check(self):
		log.debug("LED check start")
		self.threadCheck = Thread(target=self.checkLoop)
		self.threadCheck.setDaemon(True)
		self.threadCheck.start()

	def checkLoop(self):
		for l in self.leds:
			log.debug("check LED: "+l)
			self.onOld(l)
			time.sleep(5)
			self.offOld(l)
			time.sleep(1)

	def onOff(self, led, duration):
		self.onOld(led)
		t = Timer(duration, self.offOld, [led])
		t.start()

	def eyeFeedback(self):
		self.threadLED = Thread(target=self.eyeFeedbackLoop)
		self.threadLED.setDaemon(True)
		self.threadLED.start()

	def eyeFeedbackLoop(self):
		i = 1
		while i <= self.eyeFeedbackBlinks:
			self.onOff("eyeRight", self.eyeFeedbackDuration)
			self.onOff("eyeLeft", self.eyeFeedbackDuration)
			time.sleep(self.eyeFeedbackInterval)
			i += 1

	def reset(self):
		for l in self.leds:
			self.offOld(l)

	def blink(self, led):
		self.setLED(led, 1)
		time.sleep(0.001)
		self.setLED(led, 0)

	def party(self):
		log.debug("party blink start")
		self.reset()
		self.partyTimer = time.time()
		self.threadLED = Thread(target=self.partyLoop)
		self.threadLED.setDaemon(True)
		self.threadLED.start()

	def partyLoop(self):
		looping = True
		while looping:
			self.partyPattern1(4)
			self.partyPattern2()
			self.partyPattern1(5)
			self.partyPattern2()
			self.partyPattern2()
			self.partyPattern1(3)
			t = time.time() - self.partyTimer
			if t >= self.partyTime:
				looping = False
				log.debug("party blink end")
			time.sleep(self.partyInterval)

	def partyPattern1(self, t):
		for x in range(0, t+1):
			for l in self.leds:
				self.onOld(l)
			time.sleep(self.partyInterval)
			for l in self.leds:
					self.offOld(l)
			time.sleep(self.partyInterval)

	def partyPattern2(self):
		for l in self.leds:
			self.onOff(l, self.partyDuration)
			time.sleep(self.partyInterval)

	def setLED(self, led, state):
		GPIO.output(self.cathodes[led.c], state)
		GPIO.output(self.anodes[led.a], state)

	def onOld(self, led):
		self.setLEDState(led, 1)

	def offOld(self, led):
		self.setLEDState(led, 0)

	def setLEDState(self, led, state):
		if led in self.leds:
			self.leds[led].setState(state)
		else:
			log.warning("invalid led: "+led)


class LED():

	def __init__(self, c, a):
		self.a = a
		self.c = c
		self.state = 0

	def setState(self, state):
		self.state = state

